#include <iostream>
#include <verilated.h>
#include "VMESI.h"

using namespace std;

// Current simulation time (64-bit unsigned)
vluint64_t main_time = 0;
vluint64_t max_time = 200;

// Called by $time in Verilog
double sc_time_stamp() {
    return main_time;  // Note does conversion to real, to match SystemC
}

int main(int argc, char** argv, char** env){
  
  // Prevent unused variable warnings
  if (false && argc && argv && env) {}

  // Set debug level, 0 is off, 9 is highest presently used
  // May be overridden by commandArgs
  Verilated::debug(0);

  // Randomization reset policy
  // May be overridden by commandArgs
  Verilated::randReset(2);

  // Verilator must compute traced signals
  Verilated::traceEverOn(true);

  // Pass arguments so Verilated code can see them, e.g. $value$plusargs
  // This needs to be called before you create any model
  Verilated::commandArgs(argc, argv);

  // Create logs/ directory in case we have traces to put under it
  Verilated::mkdir("logs");

  // Construct the Verilated model, from Vtop.h generated from Verilating "top.v"
  VMESI* top = new VMESI();

  top->res_n = !0;
  top->clk = 0;

  top->rd = 0;
  top->wr = 0;
  top->cp = 0;
  
  top->addr = 0x80000123;

  
    
  while(!Verilated::gotFinish()){

    main_time++;

    // toggle clock
    top->clk=!top->clk;

    if(!top->clk){
    
      // Toggle control signals on an edge that doesn't correspond
      // to where the controls are sampled; in this example we do
      // this only on a negedge of clk, because we know
      // reset is not sampled there.
      if (main_time > 1 && main_time < 3) {
	top->res_n = !1;  // Assert reset
      } else {
	top->res_n = !0;  // Deassert reset
      }

      if( main_time > 3 ){

	// increment address every 20 cycles by 4
	if (0==main_time%50) {
	  top->addr += 0x4;
	}

	// alternate remote access every 20 cycles
	if (0==main_time%20) {
	  top->cp = 1;
	} else {
	  top->cp = 0;
	}

	// first half of simulation (read accesses)
	if (floor(max_time)/2>main_time) {

	  if (0==main_time%10) {
	    top->rd = 1;
	  } else {
	    top->rd = 0;
	  }
    
	} else { // second half of simulation (write accesses)	

	  if (floor(max_time/2)==main_time) {
	    top->addr = 0x80000123;
	  }	
    
	  if (0==main_time%5) {
	    top->wr = 1;
	  } else {
	    top->wr = 0;
	  }

	}

	if (max_time<=main_time){
	  top->res_n=0;
	  top->rd=1;
	  top->wr=1;
	}
	
      }

    }
       
    // Evaluate model
    // (If you have multiple models being simulated in the same
    // timestep then instead of eval(), call eval_step() on each, then
    // eval_end_step() on each.)
    top->eval();
    
    // Read outputs
    VL_PRINTF( "[%" VL_PRI64 "d] clk=%x res_n=%x addr=%04x cp=%x rd=%x wr=%x -> err=%x mesi=%x index=%x\n",
	       main_time,
	       top->clk,
	       top->res_n,
	       top->addr, top->cp, top->rd, top->wr,
	       top->err,
	       top->mesi,
	       top->index );             
  }
  
  top->final();

  //  Coverage analysis (since test passed)
#if VM_COVERAGE
  Verilated::mkdir("logs");
  VerilatedCov::write("logs/coverage.dat");
#endif

  // Delete model
  delete top;
  top = nullptr;

  // Exit
  exit(0);
  
}
